import { isNum, int } from './shims';
import { getTouch, innerWidth, innerHeight, offsetXYFromParent, outerWidth, outerHeight } from './domFns';
export function getBoundPosition(draggable, x, y) {
    // If no bounds, short-circuit and move on
    if (!draggable.props.bounds)
        return [x, y];
    // Clone new bounds
    var bounds = draggable.props.bounds;
    bounds = typeof bounds === 'string' ? bounds : cloneBounds(bounds);
    var node = findDOMNode(draggable);
    if (typeof bounds === 'string') {
        var ownerDocument = node.ownerDocument;
        var ownerWindow = ownerDocument.defaultView;
        var boundNode = void 0;
        if (bounds === 'parent') {
            boundNode = node.parentNode;
        }
        else {
            boundNode = ownerDocument.querySelector(bounds);
        }
        if (!(boundNode instanceof HTMLElement)) {
            throw new Error('Bounds selector "' + bounds + '" could not find an element.');
        }
        var nodeStyle = ownerWindow.getComputedStyle(node);
        var boundNodeStyle = ownerWindow.getComputedStyle(boundNode);
        // Compute bounds. This is a pain with padding and offsets but this gets it exactly right.
        bounds = {
            left: -node.offsetLeft + int(boundNodeStyle.paddingLeft) + int(nodeStyle.marginLeft),
            top: -node.offsetTop + int(boundNodeStyle.paddingTop) + int(nodeStyle.marginTop),
            right: innerWidth(boundNode) - outerWidth(node) - node.offsetLeft +
                int(boundNodeStyle.paddingRight) - int(nodeStyle.marginRight),
            bottom: innerHeight(boundNode) - outerHeight(node) - node.offsetTop +
                int(boundNodeStyle.paddingBottom) - int(nodeStyle.marginBottom)
        };
    }
    // Keep x and y below right and bottom limits...
    if (isNum(bounds.right))
        x = Math.min(x, bounds.right);
    if (isNum(bounds.bottom))
        y = Math.min(y, bounds.bottom);
    // But above left and top limits.
    if (isNum(bounds.left))
        x = Math.max(x, bounds.left);
    if (isNum(bounds.top))
        y = Math.max(y, bounds.top);
    return [x, y];
}
export function snapToGrid(grid, pendingX, pendingY) {
    var x = Math.round(pendingX / grid[0]) * grid[0];
    var y = Math.round(pendingY / grid[1]) * grid[1];
    return [x, y];
}
export function canDragX(props) {
    return props.axis === 'both' || props.axis === 'x';
}
export function canDragY(props) {
    return props.axis === 'both' || props.axis === 'y';
}
// Get {x, y} positions from event.
export function getControlPosition(e, touchIdentifier, draggableCore) {
    var touchObj = typeof touchIdentifier === 'number' ? getTouch(e, touchIdentifier) : null;
    if (typeof touchIdentifier === 'number' && !touchObj)
        return null; // not the right touch
    var node = findDOMNode(draggableCore);
    // User can provide an offsetParent if desired.
    var offsetParent = draggableCore.props.offsetParent || node.offsetParent || node.ownerDocument.body;
    return offsetXYFromParent(touchObj || e, offsetParent);
}
// Create an data object exposed by <DraggableCore>'s events
export function createCoreData(draggable, x, y) {
    var state = draggable.state;
    var isStart = !isNum(state.lastX);
    var node = findDOMNode(draggable);
    if (isStart) {
        // If this is our first move, use the x and y as last coords.
        return {
            node: node,
            deltaX: 0, deltaY: 0,
            lastX: x, lastY: y,
            x: x, y: y,
        };
    }
    else {
        // Otherwise calculate proper values.
        return {
            node: node,
            deltaX: x - state.lastX, deltaY: y - state.lastY,
            lastX: state.lastX, lastY: state.lastY,
            x: x, y: y,
        };
    }
}
// Create an data exposed by <Draggable>'s events
export function createDraggableData(draggable, coreData) {
    return {
        node: coreData.node,
        x: draggable.state.x + coreData.deltaX,
        y: draggable.state.y + coreData.deltaY,
        deltaX: coreData.deltaX,
        deltaY: coreData.deltaY,
        lastX: draggable.state.x,
        lastY: draggable.state.y
    };
}
// A lot faster than stringify/parse
function cloneBounds(bounds) {
    return {
        left: bounds.left,
        top: bounds.top,
        right: bounds.right,
        bottom: bounds.bottom
    };
}
function findDOMNode(draggable) {
    var node = draggable.base;
    if (!node) {
        throw new Error('<DraggableCore>: Unmounted during event!');
    }
    // $FlowIgnore we can't assert on HTMLElement due to tests... FIXME
    return node;
}
//# sourceMappingURL=positionFns.js.map