var Spring = /** @class */ (function () {
    function Spring(opt) {
        opt = opt || {};
        this.speed = opt.speed || 1 / 60.0;
        this.tension = opt.tension || 80;
        this.friction = opt.friction || 8;
        this.velocity = opt.velocity || 0;
        this.tolerance = opt.tolerance || 0.1;
        this.reset();
    }
    Spring.prototype.reset = function () {
        this.start = 0;
        this.current = this.start;
        this.end = 100;
        this.moving = true;
    };
    Spring.prototype.step = function () {
        var target = this.current;
        var before = {
            x: target - this.end,
            v: this.velocity,
            tension: this.tension,
            friction: this.friction
        };
        var after = this.integrate(before, this.speed);
        var velocity = after.v;
        var netFloat = after.x;
        var net1DVelocity = after.v;
        var netValueIsLow = Math.abs(netFloat) < this.tolerance;
        var netVelocityIsLow = Math.abs(net1DVelocity) < this.tolerance;
        var stop = netVelocityIsLow && netVelocityIsLow;
        this.current = this.end + after.x;
        this.moving = !stop;
        if (stop) {
            velocity = 0;
            this.current = this.end;
        }
        this.velocity = velocity;
        return this.current;
    };
    Spring.prototype.all = function () {
        this.reset();
        var count = 0;
        var results = [];
        while (this.moving) {
            if (count > 3000) {
                throw Error("Spring: twoo many values");
            }
            count++;
            results.push(this.step());
        }
        return results;
    };
    ;
    /**
     * Total time
     *
     * @return {Number} total time
     */
    Spring.prototype.time = function () {
        return this.all().length * this.speed;
    };
    ;
    /**
     * Integarate state
     *
     * @param {Object} state
     * @param {Number} speed
     * @return {Object}
     */
    Spring.prototype.integrate = function (state, speed) {
        var a = this.evaluate(state);
        var b = this.evaluate(state, speed * 0.5, a);
        var c = this.evaluate(state, speed * 0.5, b);
        var d = this.evaluate(state, speed, c);
        var dxdt = 1.0 / 6.0 * (a.dx + 2.0 * (b.dx + c.dx) + d.dx);
        var dvdt = 1.0 / 6.0 * (a.dv + 2.0 * (b.dv + c.dv) + d.dv);
        state.x = state.x + dxdt * speed;
        state.v = state.v + dvdt * speed;
        return state;
    };
    ;
    /**
     * Evaluate acceleration
     *
     * @param {Object} state
     * @param {Number} speed
     * @param {Object} derivative
     *
     * @return {Object}
     */
    Spring.prototype.evaluate = function (state, dt, derivative) {
        var output = {};
        var _state = {};
        if (3 === arguments.length) {
            _state.x = state.x + derivative.dx * dt;
            _state.v = state.v + derivative.dv * dt;
            _state.tension = state.tension;
            _state.friction = state.friction;
        }
        else {
            _state = state;
        }
        output.dx = _state.v;
        output.dv = this.accelerationForState(_state);
        return output;
    };
    ;
    /**
     * Calculate acceleration with setting
     *
     * @param {Object} state
     * @return {Number} acceleration
     */
    Spring.prototype.accelerationForState = function (state) {
        return -state.tension * state.x - state.friction * state.v;
    };
    ;
    return Spring;
}());
export default Spring;
// class Spring {
//   private epsilon: number;
//   private _m: number;
//   private _k: number;
//   private _c: number;
//   private _solution: number;
//   private _endPosition: number;
//   private _startTime: number;
//   constructor(mass, springConstant, damping) {
//     this.epsilon = 0.001;
//     this._m = mass;
//     this._k = springConstant;
//     this._c = damping;
//     this._solution = null;
//     this._endPosition = 0;
//     this._startTime = 0;
//     this._onComplete = () => { };
//   }
//   get springConstant() {
//     return this._k;
//   }
//   get damping() {
//     return this._c;
//   }
//   set onComplete(func) {
//     this._onComplete = func;
//   }
//   get configuration() {
//     function setSpringConstant(s, c) { s.reconfigure(1, c, s.damping()); };
//     function setSpringDamping(s, d) { s.reconfigure(1, s.springConstant(), d); }
//     return [
//       { label: 'Spring Constant', read: this.springConstant.bind(this), write: setSpringConstant.bind(this, this), min: 100, max: 1000 },
//       { label: 'Damping', read: this.damping.bind(this), write: setSpringDamping.bind(this, this), min: 1, max: 500 }
//     ];
//   }
//   almostEqual(a, b) {
//     return (a > (b - this.epsilon)) && (a < (b + this.epsilon));
//   }
//   almostZero(a) {
//     return this.almostEqual(a, 0, this.epsilon);
//   }
//   solve(initial, velocity) {
//     var c = this._c;
//     var m = this._m;
//     var k = this._k;
//     // Solve the quadratic equation; root = (-c +/- sqrt(c^2 - 4mk)) / 2m.
//     var cmk = c * c - 4 * m * k;
//     if (cmk == 0) {
//       // The spring is critically damped.
//       // x = (c1 + c2*t) * e ^(-c/2m)*t
//       var r = -c / (2 * m);
//       var c1 = initial;
//       var c2 = velocity / (r * initial);
//       return {
//         x: function (t) { return (c1 + c2 * t) * Math.pow(Math.E, r * t); },
//         dx: function (t) { var pow = Math.pow(Math.E, r * t); return r * (c1 + c2 * t) * pow + c2 * pow; }
//       };
//     } else if (cmk > 0) {
//       // The spring is overdamped; no bounces.
//       // x = c1*e^(r1*t) + c2*e^(r2t)
//       // Need to find r1 and r2, the roots, then solve c1 and c2.
//       var r1 = (-c - Math.sqrt(cmk)) / (2 * m);
//       var r2 = (-c + Math.sqrt(cmk)) / (2 * m);
//       var c2 = (velocity - r1 * initial) / (r2 - r1);
//       var c1 = initial - c2;
//       return {
//         x: function (t) { return (c1 * Math.pow(Math.E, r1 * t) + c2 * Math.pow(Math.E, r2 * t)); },
//         dx: function (t) { return (c1 * r1 * Math.pow(Math.E, r1 * t) + c2 * r2 * Math.pow(Math.E, r2 * t)); }
//       };
//     } else {
//       // The spring is underdamped, it has imaginary roots.
//       // r = -(c / 2*m) +- w*i
//       // w = sqrt(4mk - c^2) / 2m
//       // x = (e^-(c/2m)t) * (c1 * cos(wt) + c2 * sin(wt))
//       var w = Math.sqrt(4 * m * k - c * c) / (2 * m);
//       var r = -(c / 2 * m);
//       var c1 = initial;
//       var c2 = (velocity - r * initial) / w;
//       return {
//         x: function (t) { return Math.pow(Math.E, r * t) * (c1 * Math.cos(w * t) + c2 * Math.sin(w * t)); },
//         dx: function (t) {
//           var power = Math.pow(Math.E, r * t);
//           var cos = Math.cos(w * t);
//           var sin = Math.sin(w * t);
//           return power * (c2 * w * cos - c1 * w * sin) + r * power * (c2 * sin + c1 * cos);
//         }
//       };
//     }
//   }
//   x(dt) {
//     if (dt == undefined) dt = ((new Date()).getTime() - this._startTime) / 1000.0;
//     return this._solution ? this._endPosition + this._solution.x(dt) : 0;
//   }
//   dx(dt) {
//     if (dt == undefined) dt = ((new Date()).getTime() - this._startTime) / 1000.0;
//     return this._solution ? this._solution.dx(dt) : 0;
//   }
//   setEnd(x, velocity, t) {
//     if (!t) t = (new Date()).getTime();
//     if (x == this._endPosition && this.almostZero(velocity, this.epsilon)) return;
//     velocity = velocity || 0;
//     var position = this._endPosition;
//     if (this._solution) {
//       // Don't whack incoming velocity.
//       if (this.almostZero(velocity, this.epsilon))
//         velocity = this._solution.dx((t - this._startTime) / 1000.0);
//       position = this._solution.x((t - this._startTime) / 1000.0);
//       if (this.almostZero(velocity, this.epsilon)) velocity = 0;
//       if (this.almostZero(position, this.epsilon)) position = 0;
//       position += this._endPosition;
//     }
//     if (this._solution && this.almostZero(position - x, this.epsilon) && this.almostZero(velocity, this.epsilon)) {
//       return;
//     }
//     this._endPosition = x;
//     this._solution = this.solve(position - this._endPosition, velocity);
//     this._startTime = t;
//   }
//   snap(x) {
//     this._startTime = (new Date()).getTime();
//     this._endPosition = x;
//     this._solution = {
//       x: function () { return 0; },
//       dx: function () { return 0; }
//     };
//   }
//   done(t) {
//     if (!t) t = (new Date()).getTime();
//     return this.almostEqual(this.x(), this._endPosition, this.epsilon) && this.almostZero(this.dx(), this.epsilon);
//   }
//   reconfigure(mass, springConstant, damping) {
//     this._m = mass;
//     this._k = springConstant;
//     this._c = damping;
//     if (this.done()) {
//       this._onComplete();
//       return;
//     }
//     this._solution = this.solve(this.x() - this._endPosition, this.dx());
//     this._startTime = (new Date()).getTime();
//   }
//   animate(callback) {
//     const physicsModel = this;
//     function onFrame(handle, model, cb) {
//       if (handle && handle.cancelled) return;
//       cb(model.x());
//       if (!physicsModel.done() && !handle.cancelled) {
//         handle.id = requestAnimationFrame(onFrame.bind(null, handle, model, cb));
//       }
//       if (physicsModel.done()) {
//         physicsModel._onComplete();
//       }
//     }
//     function cancel(handle) {
//       if (handle && handle.id)
//         cancelAnimationFrame(handle.id);
//       if (handle)
//         handle.cancelled = true;
//       return physicsModel.x();
//     }
//     var handle = { id: 0, cancelled: false };
//     onFrame(handle, physicsModel, callback);
//     return cancel.bind(null, handle);
//   }
// }
//# sourceMappingURL=Spring.js.map